(* A Sequent Calculus for BCD Subtyping with Constructors *)

(* Coq 8.7.1 *)


Require Import List.
Require Import Permutation.
Import ListNotations.
Require Import Max.
Require Import Omega.

(* Basics on axiom of finite choice and product order over nat *)
Require Import AFC.
Require Import wf_prod.

(* Add-ons for List and Permutation *)
Require Import List_more.
Require Import Permutation_more.
Require Import Permutation_solve.


(** Generic System with Constructors **)

(** * Non-empty set of constructors with arities *)
Record Cparam := mk_Cparam {
  Ctor : Set ;
  Xc : Ctor ;
  ctvar : Ctor -> nat ;
  covar : Ctor -> nat ;
  emctx : Ctor -> bool }.



Section BCDgeneric.

Context {CP : Cparam}.

Notation α := (ctvar CP).
Notation β := (covar CP).
Notation ε := (emctx CP).


(** * Intersection types with constructors *)
Inductive itype : Set :=
| constr : Ctor CP -> list itype -> list itype -> itype
| inter : itype -> itype -> itype.


Fixpoint itype_ind_fix s : forall (P : itype -> Prop) (Pl : list itype -> Prop),
  (forall C a b, Pl a -> Pl b -> P (constr C a b)) ->
  (forall s t, P s -> P t -> P (inter s t)) ->
  (Pl nil) ->
  (forall s, P s -> forall l, Pl l -> Pl (cons s l)) ->
  P s.
Proof with try eassumption.
intros P Pl Hc Hi Hnil Hcons.
destruct s.
- apply Hc.
  + induction l...
    apply Hcons...
    eapply itype_ind_fix...
  + induction l0...
    apply Hcons...
    eapply itype_ind_fix...
- apply Hi ; eapply itype_ind_fix...
Qed.

Lemma itype_basic_ind : forall (P : itype -> Prop),
  (forall C a b, Forall P a -> Forall P b -> P (constr C a b)) ->
  (forall s t, P s -> P t -> P (inter s t)) ->
  forall s, P s.
Proof.
intros.
eapply itype_ind_fix ; try eassumption ; constructor ; assumption.
Qed.


Notation "s ∩ t" := (inter s t) (at level 40).
Notation "C @ a , b " := (constr C a b) (at level 40).
Definition X := (Xc CP)@nil,nil.



(** * Appropriate use of arities *) 
Inductive ivalid : itype -> Prop :=
| iv_c : forall C a b, length a = α C -> length b = β C ->
           Forall ivalid a -> Forall ivalid b -> ivalid (C@a,b)
| iv_i : forall s t, ivalid s -> ivalid t -> ivalid (s ∩ t).

Definition lvalid C L :=
  Forall (fun c => length (fst c) = α C
                /\ length (snd c) = β C
                /\ Forall ivalid (fst c)
                /\ Forall ivalid (snd c)) L.

Lemma lvalid_ivalid : forall C L, lvalid C L ->
  Forall ivalid (map (fun x => (C@fst x, snd x)) L).
Proof with try assumption.
induction L ; intros Hv ; constructor. 
- inversion Hv ; subst.
  destruct H1 as (H11 & H12 & H13 & H14).
  constructor...
- apply IHL.
  inversion Hv...
Qed.


(** * Bound on size of types *) 
Inductive isize : nat -> itype -> Prop :=
| is_c : forall C a b m n, Forall (isize m) a -> Forall (isize m) b ->
           m < n -> isize n (C@a,b)
| is_i : forall s t n, isize n s -> isize n t -> isize (S n) (s ∩ t).

Lemma isize_inc : forall s n m, isize n s -> n <= m -> isize m s.
Proof with try assumption ; try omega.
apply (itype_basic_ind
        (fun x => forall n m, isize n x -> n <= m -> isize m x)).
- intros C a b Ha Hb n m Hn Hl.
  inversion Hn ; subst.
  apply is_c with m0...
- intros s t Hs Ht n m Hn Hl.
  inversion Hn ; subst.
  inversion Hl ; subst...
  apply (Hs _ m0) in H2...
  apply (Ht _ m0) in H3...
  apply is_i...
Qed.

Lemma exists_isize : forall s, exists n, isize n s.
Proof with try eassumption ; try (zify ; omega).
apply itype_basic_ind.
- intros C a b Ha Hb.
  assert (Hinc := isize_inc).
  apply AFClist in Ha ;
    try (intros ; eapply Hinc ; eassumption ; fail)...
  apply AFClist in Hb ;
    try (intros ; eapply Hinc ; eassumption ; fail)...
  destruct Ha as [ka Ha].
  destruct Hb as [kb Hb].
  exists (S (max ka kb)).
  apply is_c with (max ka kb)...
  + induction a ; constructor ; inversion Ha ; subst.
    * eapply Hinc...
    * apply IHa...
  + induction b ; constructor ; inversion Hb ; subst.
    * eapply Hinc...
    * apply IHb...
- intros s t Hs Ht.
  destruct Hs as [n1 Hs].
  destruct Ht as [n2 Ht].
  exists (S (max n1 n2)).
  apply is_i ; eapply isize_inc...
Qed.

(** * Projections for triples *)
Definition scd {A B C} : A * (B * C) -> B := fun x => fst (snd x).
Definition trd {A B C} : A * (B * C) -> C := fun x => snd (snd x).


(** * Subtyping derivation system with size parameter *)
Inductive subtype : list itype -> itype -> nat -> Prop :=
| wk : forall C a b G D s n, subtype (G ++ D) s n ->
         ivalid (C@a,b) -> subtype (G ++ C@a,b :: D) s (S n)
| ir : forall G s t n, subtype G s n -> subtype G t n ->
         subtype G (s ∩ t) (S n)
| il : forall G D s t u n, subtype (G ++ s :: t :: D) u n ->
         subtype (G ++ s ∩ t :: D) u (S n)
| ct : forall L C a b n,
         (length L = 0 -> ε C = true /\ Forall ivalid a) ->
         (forall i, i < α C ->
            Forall
            (fun c => subtype [nth i a X] (nth i (fst c) X) n) L) ->
         (forall i, i < β C ->
            subtype (map (fun c => (nth i (snd c) X)) L)
                    (nth i b X) n) ->
         lvalid C L -> length a = α C -> length b = β C ->
         subtype (map (fun c => C@fst c,snd c) L) (C@a,b) (S n).

Lemma subtype_inc : forall n m G s,
  subtype G s n -> n <= m -> subtype G s m.
Proof with try omega ; try assumption.
induction n using (well_founded_induction lt_wf).
intros m G s Hm Hl.
inversion Hm ; subst.
- intros.
  inversion Hl ; apply wk...
  apply (H n0)...
- inversion Hl ; apply ir...
  + apply (H n0)...
  + apply (H n0)...
- inversion Hl ; apply il...
  apply (H n0)...
- inversion Hl ; apply ct...
  + intros i Hi.
    assert (HF := H1 i Hi).
    clear H0 H1 H2 H3 H4 H5 H6 Hm.
    induction L ; constructor ; inversion HF.
    * apply (H n0)...
    * apply IHL...
  + intros i Hi.
    assert (Hs := H2 i Hi).
    clear H0 H1 H2 H3 H4 H5 H6 Hm.
    apply (H n0)...
Qed.

Lemma subtype_ivalid :
  forall n G s, subtype G s n -> Forall ivalid G /\ ivalid s.
Proof with try omega ; try eassumption.
induction n using (well_founded_induction lt_wf) ;
  intros G s Hsub ; inversion Hsub ; subst ; split ;
  try (eapply (H n0) ; try omega ; try eassumption ; fail).
- apply H in H0...
  destruct H0 as [HGD _].
  apply Forall_app_inv in HGD.
  destruct HGD as [HG HD].
  apply Forall_app ; [ | constructor ]...
- constructor ; eapply (H n0)...
- apply H in H0...
  destruct H0 as [H0 _].
  apply Forall_app_inv in H0.
  destruct H0 as [HG HD].
  inversion HD ; subst.
  inversion H3 ; subst.
  apply Forall_app ; [ | constructor ]...
  constructor...
- apply lvalid_ivalid...
- constructor...
  + rewrite <- H4 in H1.
    destruct L.
    * simpl in H0.
      assert (Hv := proj2 (H0 eq_refl))...
    * assert (forall i, i < length a -> exists s, subtype [nth i a X] s n0)
        as Ha.
      { intros i Hi.
        assert (Hs := H1 i Hi).
        inversion Hs.
        eexists... }
      revert Ha ; clear - H0 H ; induction a ; intros Ha ; constructor.
      -- assert (0 < length (a :: a0)) as Hs by (simpl ; omega).
         apply Ha in Hs.
         inversion Hs...
         apply H in H1...
         destruct H1 as [H1 _].
         inversion H1...
      -- apply IHa.
         intros HL ; inversion HL.
         intros.
         assert (HaS := Ha (S i))...
         simpl in HaS.
         apply HaS...
  + rewrite <- H5 in H2.
    assert (forall i, i < length b -> ivalid (nth i b X))
      as Hb.
    { intros i Hi.
      assert (Hs := H2 i Hi).
      apply H in Hs...
      apply Hs. }
    revert Hb ; clear ; induction b ; intros Hb ; constructor.
    * assert (0 < length (a :: b)) as Hs by (simpl ; omega).
      apply Hb in Hs.
      inversion Hs...
    * apply IHb.
      intros.
      assert (HbS := Hb (S i))...
      simpl in HbS.
      apply HbS...
Qed.


(** * Admissible rules *)
Lemma ex : forall D G s n,
  subtype G s n -> Permutation G D -> subtype D s n.
Proof with try eassumption.
intros D G s n H.
revert D ; induction H ; intros G' HP.
- assert (Heq := HP).
  symmetry in Heq.
  apply Permutation_vs_elt_inv in Heq.
  destruct Heq as (l' & l'' & Heq) ; subst.
  apply Permutation_app_inv in HP.
  apply IHsubtype in HP.
  apply wk...
- assert (HP2 := HP).
  apply IHsubtype1 in HP.
  apply IHsubtype2 in HP2.
  apply ir...
- assert (Heq := HP).
  symmetry in Heq.
  apply Permutation_vs_elt_inv in Heq.
  destruct Heq as (l' & l'' & Heq) ; subst.
  apply Permutation_app_inv in HP.
  apply (Permutation_elt t) in HP.
  apply (Permutation_elt s) in HP.
  apply IHsubtype in HP.
  apply il...
- symmetry in HP.
  apply Permutation_map_inv in HP.
  destruct HP as (L' & Heq & HP) ; subst.
  apply ct...
  + intro HL'.
    destruct L' ; inversion HL'.
    symmetry in HP.
    apply Permutation_nil in HP ; subst.
    apply H...
  + intros i Hi.
    apply H0 in Hi.
    eapply Permutation_Forall...
  + intros i Hi.
    apply H2...
    apply Permutation_map...
  + unfold lvalid in H3.
    unfold lvalid.
    eapply Permutation_Forall...
Qed.

Lemma wk_gen : forall t, ivalid t -> forall G D s n,
  subtype (G ++ D) s n -> exists m, subtype (G ++ t :: D) s m.
Proof with try eassumption.
induction t ; intros.
- inversion H.
  eexists.
  apply wk...
- inversion H ; subst.
  destruct (IHt2 H4 _ _ _ _ H0) as [m2 H2].
  destruct (IHt1 H3 _ _ _ _ H2) as [m1 H1].
  eexists.
  apply il...
Qed.

Lemma ax_size : forall n s,
  isize n s -> ivalid s -> exists m, subtype [s] s m.
Proof with try eassumption ; try (zify ; omega).
induction n using (well_founded_induction lt_wf).
intros s Hsize Hs.
inversion Hsize ; subst.
- assert (forall i, i < length a -> isize m (nth i a X)) as Hsizea.
  { clear Hs Hsize.
    induction H0 ; intros i Hi.
    - inversion Hi.
    - destruct i ; simpl...
      apply IHForall...
      inversion Hi... }
  assert (forall i, i < length b -> isize m (nth i b X)) as Hsizeb.
  { clear Hs Hsize.
    induction H1 ; intros i Hi.
    - inversion Hi.
    - destruct i ; simpl...
      apply IHForall...
      inversion Hi... }
  assert
    (forall i, i < α C -> exists n,
       Forall (fun c =>
         subtype [nth i a X] (nth i (scd c) X) n) [(C,(a,b))])
  as Ha.
  { intros i Hi.
    inversion Hs ; subst ; simpl.
    destruct (H m H2 (nth i a X)).
    - apply Hsizea...
    - rewrite <- H6 in Hi.
      revert i Hi H8 ; clear.
      induction a ; intros i Hi Hl.
      + inversion Hi.
      + inversion Hl ; subst.
        destruct i ; simpl...
        apply IHa...
        inversion Hi...
    - eexists.
      constructor ; try (constructor ; fail)... }
  assert
    (forall i, i < β C -> exists n,
       subtype [nth i b X] (nth i b X) n)
  as Hb.
  { intros i Hi.
    inversion Hs ; subst ; simpl.
    destruct (H m H2 (nth i b X)).
    - apply Hsizeb...
    - rewrite <- H7 in Hi.
      revert i Hi H9 ; clear.
      induction b ; intros i Hi Hl.
      + inversion Hi.
      + inversion Hl ; subst.
        destruct i ; simpl...
        apply IHb...
        inversion Hi...
    - eexists... }
  apply AFCinc in Ha ; [ apply AFCinc in Hb | ].
  + destruct Ha as [ka Ha].
    destruct Hb as [kb Hb].
    eexists.
    change [C@a,b] with (map (fun c => C @ fst c, snd c) [(a,b)]).
    apply (ct _ _ _ _ (max ka kb))...
    * intro Hl ; inversion Hl.
    * intros i Hi.
      constructor ; try (constructor ; fail).
      apply (subtype_inc ka)...
      assert (Ha' := Ha i Hi).
      inversion Ha'...
    * intros i Hi.
      apply (subtype_inc kb)...
      apply Hb...
    * inversion Hs ; subst.
      constructor ; try (constructor ; fail).
      split ; [ | split ; [ | split]]...
    * inversion Hs...
    * inversion Hs...
  + intros p i j Hi Hl.
    eapply subtype_inc...
  + intros p i j Hi Hl.
    inversion Hi ; subst.
    constructor ; try (constructor ; fail).
    eapply subtype_inc...
- inversion Hs ; subst.
  assert (n0 < S n0) as Hl by omega.
  destruct (H n0 Hl s0 H0 H4) as [k1 Hs1].
  destruct (H n0 Hl t  H1 H5) as [k2 Hs2].
  rewrite <- (app_nil_r [s0]) in Hs1.
  apply (wk_gen t) in Hs1...
  destruct Hs1 as [k1' Hs1].
  rewrite <- (app_nil_l [t]) in Hs2.
  apply (wk_gen s0) in Hs2...
  destruct Hs2 as [k2' Hs2].
  exists (S (S (S (max k1' k2')))).
  rewrite <- (app_nil_l [s0 ∩ t]).
  apply il.
  apply ir ; eapply subtype_inc...
Qed.

Lemma ax : forall s, ivalid s -> exists n, subtype [s] s n.
Proof.
intro s.
destruct (exists_isize s) as [n Hs].
eapply ax_size ; eassumption.
Qed.

Lemma il_inv : forall n G D s t u, subtype (G ++ s ∩ t :: D) u n ->
  exists m, m < n /\ subtype (G ++ s :: t :: D) u m.
Proof with try eassumption ; try (zify ; omega).
induction n using (well_founded_induction lt_wf).
intros G D s t u Hi.
inversion Hi ; subst.
- dichot_elt_app_exec H0 ; subst.
  + rewrite app_assoc in H1.
    apply H in H1...
    destruct H1 as (s' & Hm & H1).
    rewrite <- app_assoc in H1.
    exists (S s') ; split...
    rewrite <- app_assoc.
    apply wk...
  + destruct l0 ; inversion H4 ; subst.
    rewrite <- ? app_assoc in H1.
    rewrite <- ? app_comm_cons in H1.
    apply H in H1...
    destruct H1 as (s' & Hm & H1).
    rewrite ? app_comm_cons in H1.
    rewrite app_assoc in H1.
    exists (S s') ; split...
    rewrite ? app_comm_cons.
    rewrite app_assoc.
    apply wk...
- eapply H in H0...
  eapply H in H1...
  destruct H0 as (nl & Hml & Hl).
  destruct H1 as (nr & Hmr & Hr).
  exists (S (max nl nr)) ; split...
  apply ir ; eapply subtype_inc...
- dichot_elt_app_exec H0 ; subst.
  + rewrite ? app_comm_cons in H1.
    rewrite app_assoc in H1.
    apply H in H1...
    destruct H1 as (s' & Hm & H1).
    rewrite <- ? app_assoc in H1.
    rewrite <- ? app_comm_cons in H1.
    exists (S s') ; split...
    rewrite <- ? app_assoc.
    rewrite <- ? app_comm_cons.
    apply il...
  + destruct l0 ; inversion H3 ; subst.
    * rewrite app_nil_r in H1.
      eexists ; split...
    * rewrite <- ? app_assoc in H1.
      rewrite <- ? app_comm_cons in H1.
      apply H in H1...
      destruct H1 as (s' & Hm & H1).
      rewrite ? app_comm_cons in H1.
      rewrite app_assoc in H1.
      exists (S s') ; split...
      rewrite ? app_comm_cons.
      rewrite app_assoc.
      apply il...
- exfalso.
  clear - H0 ; revert G H0 ; induction L ;
    intros G H ; destruct G ; inversion H.
  apply IHL in H2...
Qed.


Lemma ctr_size : forall v s, isize (fst v) s ->
forall G u, subtype (s :: s :: G) u (snd v) ->
  exists m, subtype (s :: G) u m.
Proof with try lt_nat_nat_solve ;
           try eassumption ;
           try reflexivity ; try (zify ; omega).
induction v as [(f,n) IH] using (well_founded_induction wf_nat_nat).
simpl ; intros s Hs G u pi.
inversion pi ; subst.
- destruct G0 ; [ | destruct G0 ] ; inversion H ; subst.
  + eexists...
  + eexists...
  + rewrite <- ? app_comm_cons in H0.
    apply (IH (f,n0)) in H0...
    destruct H0 as [m H0].
    eexists...
    rewrite app_comm_cons.
    apply wk...
- apply (IH (f,n0)) in H...
  destruct H as [m1 H].
  apply (IH (f,n0)) in H0...
  destruct H0 as [m2 H0].
  exists (S (max m1 m2)).
  apply ir ; eapply subtype_inc ; try eassumption...
- destruct G0 ; [ | destruct G0 ] ; inversion H ; subst.
  + assert (Permutation ([] ++ s0 :: t :: s0 ∩ t :: G)
                        ((s0 :: nil) ++ s0 ∩ t :: t :: G))
      as HP by perm_solve.
    eapply ex in HP...
    apply il_inv in HP.
    destruct HP as (m & _ & Hsub).
    simpl in Hsub.
    inversion Hs ; subst.
    apply (IH (n,m)) in Hsub...
    destruct Hsub as [m1 Hsub].
    assert (Permutation (s0 :: t :: t :: G) (t :: t :: s0 :: G))
      as HP by perm_solve.
    eapply ex in HP...
    apply (IH (n,m1)) in HP...
    destruct HP as [m2 Hsub2].
    eapply ex in Hsub2 ; [ | apply perm_swap ].
    eexists.
    rewrite <- (app_nil_l (_ :: G)).
    apply il...
  + simpl in H0.
    rewrite <- (app_nil_l (_ :: _)) in H0.
    apply il_inv in H0.
    destruct H0 as (m1 & _ & H0).
    inversion Hs ; subst.
    assert (Permutation ([] ++ s0 :: t :: s0 :: t :: G)
                        (s0 :: s0 :: t :: t :: G))
      as HP by perm_solve.
    eapply (ex _ _ _ _ H0) in HP.
    apply (IH (n,m1)) in HP...
    destruct HP as [m2 Hsub].
    assert (Permutation (s0 :: t :: t :: G) (t :: t :: s0 :: G))
      as HP by perm_solve.
    eapply ex in HP...
    apply (IH (n,m2)) in HP...
    destruct HP as [m3 Hsub2].
    eapply ex in Hsub2 ; [ | apply perm_swap ].
    eexists.
    rewrite <- (app_nil_l (_ :: G)).
    apply il...
  + rewrite <- ? app_comm_cons in H0.
    apply (IH (f,n0)) in H0...
    destruct H0 as [m H0].
    eexists...
    rewrite app_comm_cons.
    apply il...
- symmetry in H.
  decomp_map H ; subst.
  inversion H9.
  assert (forall i, i < β C -> exists n1,
          subtype (map (fun c => nth i (snd c) X) (x0 :: l1)) (nth i b X) n1)
    as Hb.
  { intros i Hi.
    assert (HI := H2 i Hi).
    simpl in HI.
    change n0 with (snd (f,n0)) in HI.
    rewrite H7 in HI.
    apply IH in HI...
    simpl ; rewrite <- H7.
    simpl in Hs ; inversion Hs ; subst.
    apply Forall_nth...
    - remember (snd x) as l'.
      revert H13 H14 ; clear ; induction l' ; intros Hs Hl ;
        inversion Hs ; subst ; constructor ; 
        [ eapply isize_inc | apply IHl' ]...
    - inversion H3.
      destruct H10 as [_ [Hlen _]].
      rewrite Hlen... }
  apply AFCinc in Hb ; [ | intros ; eapply subtype_inc ]...
  destruct Hb as [k Hb].
  eexists (S (max n0 k)).
  rewrite H7 ; rewrite H6.
  change (C @ fst x0, snd x0
           :: map (fun c => C @ fst c, snd c) l1)
    with (map (fun c => C @ fst c, snd c) (x0::l1)).
  apply ct...
  + intro Hl ; inversion Hl.
  + intros i Hi.
    assert (H1' := H1 i Hi).
    inversion H1'.
    inversion H11.
    constructor.
    * eapply subtype_inc...
    * revert H15 ; clear ; induction l1 ; intro H ; constructor ; inversion H ;
        [ eapply subtype_inc | apply IHl1]...
  + intros i Hi.
    apply Hb in Hi.
    eapply subtype_inc...
  + inversion H3...
Qed.

Lemma ctr : forall G s u n, subtype (s :: s :: G) u n ->
  exists m, subtype (s :: G) u m.
Proof with try eassumption.
intros.
assert (Hs := exists_isize s).
destruct Hs as [f Hs].
change f with (fst (f,n)) in Hs.
change n with (snd (f,n)) in H.
apply ctr_size in H...
Qed.


Theorem cut_size : forall v s, isize (fst v) s ->
forall D G t, subtype D s (scd v) -> subtype (s :: G) t (trd v) ->
  exists m, subtype (D ++ G) t m.
Proof with try lt_nat_nat_nat_solve ;
           try eassumption ; try reflexivity ; try (zify ; omega) ; try perm_solve.
induction v as [(f,(l,r)) IH] using (well_founded_induction wf_nat_nat_nat).
intros s Hs D G t pi1 pi2.
inversion pi1 ; subst.
- unfold scd in H3 ; simpl in H3 ; subst.
  change n with (scd (f,(n,r))) in H.
  eapply IH in H...
  destruct H as [m H].
  eexists.
  rewrite <- ? app_assoc in H.
  rewrite <- ? app_assoc ; rewrite <- app_comm_cons.
  apply wk...
- inversion pi2 ; subst.
  + unfold scd in H3 ; simpl in H3 ; subst.
    unfold trd in H6 ; simpl in H6 ; subst.
    destruct G0 ; inversion H1 ; subst.
    rewrite <- app_comm_cons in H2.
    change n0 with (trd (f, (S n, n0))) in H2.
    change (scd (f, (S n, S n0))) with (scd (f, (S n, n0))) in pi1.
    eapply IH in H2...
    destruct H2 as [m HI].
    rewrite app_assoc in HI.
    eexists.
    rewrite app_assoc.
    apply wk...
  + unfold scd in H3 ; simpl in H3 ; subst.
    unfold trd in H6 ; simpl in H6 ; subst.
    change n0 with (trd (f,(S n,n0))) in H1.
    change n0 with (trd (f,(S n,n0))) in H2.
    change (scd (f,(S n, S n0))) with (scd (f,(S n,n0))) in pi1.
    eapply IH in H1...
    eapply IH in H2...
    destruct H1 as [m1 HI1].
    destruct H2 as [m2 HI2].
    eexists (S (max m1 m2)).
    apply ir ; eapply subtype_inc...
  + unfold scd in H3 ; simpl in H3 ; subst.
    unfold trd in H5 ; simpl in H5 ; subst.
    inversion Hs ; subst.
    destruct G0 ; inversion H1 ; subst.
    * simpl in H5 ; subst.
      change n0 with (trd (n1, (n, n0))) in H2.
      change n  with (scd (n1, (n, n0))) in H.
      eapply IH in H2...
      destruct H2 as [m H2].
      change n with (scd (n1, (n, m))) in H0.
      change m with (trd (n1, (n, m))) in H2.
      eapply ex in H2 ; [ | symmetry ; apply Permutation_cons_app ; reflexivity ].
      eapply IH in H2...
      destruct H2 as [m' HI].
      clear - HI ; revert m' G HI ; induction D ; intros m G HI.
      -- eexists...
      -- rewrite <- ? app_comm_cons in HI.
         assert (Permutation (a :: D ++ a :: D ++ G) (a :: a :: D ++ D ++ G))
           as HP by perm_solve.
         apply (ex _ _ _ _ HI) in HP...
         apply ctr in HP.
         destruct HP as [m1 HP].
         eapply ex in HP.
         ++ apply IHD in HP.
            destruct HP as [m2 HP].
            eexists.
            rewrite <- ? app_comm_cons.
            eapply ex ; [ | symmetry ; apply Permutation_cons_app ]...
         ++ perm_solve.
    * change (scd (f, (S n, S n0))) with (scd (f, (S n, n0))) in pi1.
      change n0 with (trd (f,(S n,n0))) in H2.
      rewrite <- app_comm_cons in H2.
      eapply IH in H2...
      destruct H2 as [m' H2].
      rewrite app_assoc in H2.
      eexists.
      rewrite app_assoc.
      apply il...
  + exfalso.
    symmetry in H1 ; decomp_map H1.
    inversion H11.
- unfold scd in H2 ; simpl in H2 ; subst.
  change n with (scd (f, (n, r))) in H.
  eapply IH in H...
  destruct H as [m H].
  eexists.
  rewrite <- app_assoc in H ; rewrite <- ? app_comm_cons in H.
  rewrite <- app_assoc ; rewrite <- app_comm_cons.
  apply il...
- inversion pi2 ; subst.
  + unfold scd in H7 ; simpl in H7 ; subst.
    unfold trd in H10 ; simpl in H10 ; subst.
    destruct G0 ; inversion H5 ; subst.
    * revert H6 ; clear - H2 ; induction L ; intro HI.
      -- eexists ; apply HI.
      -- inversion H2 ; subst.
         apply IHL in HI...
         destruct HI as [m HI].
         eexists.
         simpl ; rewrite <- (app_nil_l (constr _ _ _ :: _)).
         apply wk...
         apply lvalid_ivalid in H2.
         inversion H2...
    * rewrite <- app_comm_cons in H6.
      change n0 with (trd (f,(S n,n0))) in H6.
      change (scd (f,(S n, S n0))) with (scd (f,(S n,n0))) in pi1.
      eapply IH in H6...
      destruct H6 as [m HI].
      rewrite app_assoc in HI.
      eexists.
      rewrite app_assoc.
      apply wk...
  + unfold scd in H7 ; simpl in H7 ; subst.
    unfold trd in H10 ; simpl in H10 ; subst.
    change n0 with (trd (f, (S n, n0))) in H5.
    change n0 with (trd (f, (S n, n0))) in H6.
    change (scd (f, (S n, S n0))) with (scd (f, (S n,n0))) in pi1.
    eapply IH in H5...
    eapply IH in H6...
    destruct H5 as [m1 HI1].
    destruct H6 as [m2 HI2].
    eexists (S (max m1 m2)).
    apply ir ; eapply subtype_inc...
  + unfold scd in H7 ; simpl in H7 ; subst.
    unfold trd in H9 ; simpl in H9 ; subst.
    destruct G0 ; inversion H5 ; subst.
    change n0 with (trd (f, (S n, n0))) in H6.
    change (scd (f, (S n, S n0))) with (scd (f, (S n, n0))) in pi1.
    rewrite <- app_comm_cons in H6.
    eapply IH in H6...
    destruct H6 as [m HI].
    rewrite app_assoc in HI.
    eexists.
    rewrite app_assoc.
    apply il...
  + unfold scd in H7 ; simpl in H7 ; subst.
    unfold trd in H14 ; simpl in H14 ; subst.
    symmetry in H5 ; decomp_map H5 ; inversion H13 ; subst.
    assert (forall i : nat, i < α C0 -> exists k, 
            Forall (fun c => subtype [nth i a0 X] (nth i (fst c) X) k)
              (L ++ l0)) as Ha.
    { intros i Hi.
      apply AFClist ; [ intros ; eapply subtype_inc ; eassumption | ].
      assert (Hl2 := H8 i Hi).
      inversion Hl2 ; subst.
      apply Forall_app.
      - assert (HL := H0 i Hi).
        revert HL ; clear - IH Hs H10 Hi H14 ; induction L ; intro HL ;
          constructor ; inversion HL ; subst.
        + inversion Hs ; subst.
          simpl in H7.
          change n0 with (scd (m, (n0, n))) in H14.
          change n  with (trd (m, (n0, n))) in H2.
          eapply IH in H14...
          * assumption.
          * apply Forall_nth...
            inversion H10...
        + apply IHL...
      - apply exists_Forall.
        eexists... }
    apply AFCinc in Ha.
    * destruct Ha as [ka Ha].
      assert (forall i : nat, i < β C0 -> exists k,
                subtype (map (fun c => nth i (snd c) X) (L ++ l0))
                        (nth i b0 X) k) as Hb.
      { intros i Hi.
        assert (HL := H1 i Hi).
        assert (HR := H9 i Hi).
        inversion Hs ; subst.
        simpl in H18.
        change n  with (scd (m, (n, n0))) in HL.
        change n0 with (trd (m, (n, n0))) in HR.
        rewrite ? map_app.
        eapply IH in HL...
        apply Forall_nth... }
      apply AFCinc in Hb ; [ | intros ; eapply subtype_inc ; eassumption ].
      destruct Hb as [kb Hb].
      exists (S (max (max n n0) (max ka kb))).
      rewrite <- map_app.
      apply ct...
      -- intro Hl.
         rewrite app_length in Hl.
         split.
         ++ apply H...
         ++ assert (l0 = []) as Hl0 ; subst.
            { destruct l0...
              simpl in Hl... }
            rewrite <- H11 in H8.
            assert (forall i, i < length a0 -> ivalid (nth i a0 X)) as Hv.
            { intros i Hi.
              apply H8 in Hi.
              inversion Hi.
              apply subtype_ivalid in H14...
              destruct H14 as [H14 _].
              inversion H14... }
            revert Hv ; clear ; induction a0 ; intros Hv ; constructor.
            ** assert (0 < length (a :: a0)) as Ha0 by (simpl ; omega).
               apply Hv in Ha0...
            ** apply IHa0.
               intros i Hi.
               assert (S i < length (a :: a0)) as HaS by (simpl ; omega).
               apply Hv in HaS...
      -- intros i0 Hi0.
         assert (Ha' := Ha i0 Hi0).
         eapply (inc_Forall
                  (fun i c => subtype [nth i0 a0 X] (nth i0 (fst c) X) i))...
         intros ; eapply subtype_inc...
      -- intros i0 Hi0.
         assert (Hb' := Hb i0 Hi0).
         eapply subtype_inc...
      -- clear - H2 H10.
         revert H2 ; induction L ; intro HL...
         ++ inversion H10...
         ++ inversion HL.
            constructor...
            apply IHL...
    * intros.
      eapply (inc_Forall
               (fun i c => subtype [nth n1 a0 X] (nth n1 (fst c) X) i))...
      intros ; eapply subtype_inc...
Qed.

Theorem cut : forall D s G t n m,
  subtype D s n -> subtype (s :: G) t m -> exists k, subtype (D ++ G) t k.
Proof.
intros.
assert (Hs := exists_isize s).
destruct Hs as [f Hs].
change f with (fst (f,(n,m))) in Hs.
change n with (scd (f,(n,m))) in H.
change m with (trd (f,(n,m))) in H0.
eapply cut_size ; eassumption.
Qed.

End BCDgeneric.





(** BCD presentation *)
Section BCDstd.

Variable BCtor : Set.
Variable bvar : BCtor -> nat.

Definition BCDCtor := (option (option BCtor)).
  (* None is Omega, Some None is Arrow *)
Fixpoint bctvar (C:BCDCtor) :=
  match C with
  | Some None => 1
  | _ => 0
  end.
Fixpoint bcovar C :=
  match C with
  | None => 0
  | Some None => 1
  | Some (Some c) => bvar c
  end.
Fixpoint bemctx (C:BCDCtor) :=
  match C with
  | None => true
  | Some None => true
  | _ => false
  end.

Proposition BCDC : Cparam.
Proof with try reflexivity ; try assumption ; try omega.
apply (mk_Cparam BCDCtor None bctvar bcovar bemctx).
Defined.


Notation α := (ctvar BCDC).
Notation β := (covar BCDC).
Notation ε := (emctx BCDC).
Notation "s ∩ t" := (inter s t) (at level 40).
Notation "C @ a , b " := (@constr BCDC C a b) (at level 40).
Notation Ω := (None@nil,nil).
Notation "s → t" := ((Some None)@[s],[t]) (at level 42).
Notation κ := (fun c a => ((Some (Some c)) @ [] , a)).

Reserved Notation "a ≤ b " (at level 70).
Inductive sub_bcd : @itype BCDC -> @itype BCDC -> Prop :=
| id_bcd          : forall a, a ≤ a
| trans_bcd       : forall a b c, a ≤ b -> b ≤ c -> ivalid b -> a ≤ c
| omega_right_bcd : forall a, a ≤ Ω
| inter_left1_bcd : forall a b, a ∩ b ≤ a
| inter_left2_bcd : forall a b, a ∩ b ≤ b
| inter_right_bcd : forall a, a ≤ a ∩ a
| inter_bcd       : forall a b c d, a ≤ c -> b ≤ d -> a ∩ b ≤ c ∩ d
| arrow_bcd       : forall a b c d, c ≤ a -> b ≤ d -> a → b ≤ c → d
| arrow_inter_bcd : forall a b c, (a → b) ∩ (a → c) ≤ a → b ∩ c
| arrow_omega_bcd : Ω ≤ Ω → Ω
| ctor_mon_bcd    : forall C l1 l2 a b, a ≤ b -> κ C (l1++a::l2) ≤ κ C (l1++b::l2)
| ctor_inter_bcd  : forall C l1 l2, κ C l1 ∩ κ C l2 ≤ κ C (map2 inter l1 l2)
where "a ≤ b" := (sub_bcd a b).

Lemma ctor_mon_bcd_gen : forall C l1 l2, ivalid (κ C l1) -> ivalid (κ C l2) ->
  Forall2 (fun s t => s ≤ t) l1 l2 -> κ C l1 ≤ κ C l2.
Proof with try omega ; try eassumption.
intro C.
assert (forall k l1 l2 l, k = length l1 -> k = length l2 ->
          ivalid (κ C (l++l1)) -> ivalid (κ C (l++l2)) ->
          Forall2 (fun s t => s ≤ t) l1 l2 -> κ C (l++l1) ≤ κ C (l++l2))
  as Hlength.
{ induction k ; intros l1 l2 l Hl1 Hl2 Hv1 Hv2 HF ; inversion HF ; subst.
  - apply id_bcd.
  - inversion Hl1.
  - inversion Hl1.
  - assert (Forall ivalid (l++y::l0)) as Hv0.
    { inversion Hv2.
      apply Forall_app_inv in H7.
      destruct H7.
      inversion H8.
      inversion Hv1.
      apply Forall_app_inv in H19.
      destruct H19.
      inversion H20.
      apply Forall_app...
      constructor... }
    eapply trans_bcd.
    + apply (ctor_mon_bcd _ l)...
    + replace (l ++ y :: l0) with ((l ++ y :: []) ++ l0)
        by (rewrite <- app_assoc ; reflexivity).
      replace (l ++ y :: l') with ((l ++ y :: []) ++ l')
        by (rewrite <- app_assoc ; reflexivity).
      apply IHk...
      * simpl in Hl1...
      * simpl in Hl2...
      * rewrite <- app_assoc.
        rewrite <- app_comm_cons.
        rewrite app_nil_l...
        constructor ; try constructor...
        inversion Hv1.
        rewrite app_length in H5.
        simpl in H5.
        rewrite ? app_length...
      * rewrite <- app_assoc...
    + constructor ; try constructor...
      inversion Hv1.
      rewrite app_length in H5.
      simpl in H5.
      rewrite ? app_length... }
intros l1 l2 Hl1 Hl2.
rewrite <- (app_nil_l l1).
rewrite <- (app_nil_l l2).
apply (Hlength (length l1))...
inversion Hl1 ; inversion Hl2...
Qed.

(** * Correctness with respect to BCD *)
Theorem correctness : forall a b, ivalid a -> ivalid b ->
  a ≤ b -> exists n, subtype [a] b n.
Proof with try eassumption ; try reflexivity ; try (zify ; omega).
intros a b Hva Hvb Hbcd.
induction Hbcd.
- apply ax...
- apply IHHbcd1 in Hva...
  apply IHHbcd2 in Hvb...
  destruct Hva as [n Hl].
  destruct Hvb as [m Hr].
  rewrite <- (app_nil_r [a]).
  eapply cut...
- rewrite <- (app_nil_l [a]).
  eapply wk_gen...
  change [] with (map (fun x => (None@fst x,snd x)) []).
  apply ct ; try constructor...
  + constructor.
  + intros i Hi.
    inversion Hi.
  Unshelve. exact 0.
- inversion Hva.
  assert (Hax := ax a Hvb).
  destruct Hax as [n Hax].
  rewrite <- (app_nil_l [_]) in Hax.
  eapply (wk_gen b) in Hax...
  destruct Hax as [m Hax].
  eexists.
  rewrite <- (app_nil_l [_]).
  apply il.
  apply (ex _ [b;a])...
  apply Permutation_rev.
- inversion Hva.
  assert (Hax := ax b H2).
  destruct Hax as [n Hax].
  rewrite <- (app_nil_l [_]) in Hax.
  eapply (wk_gen a) in Hax...
  destruct Hax as [m Hax].
  eexists.
  rewrite <- (app_nil_l [_]).
  apply il...
- assert (Hax := ax a Hva).
  destruct Hax as [n Hax].
  eexists.
  apply ir...
- inversion Hva.
  inversion Hvb.
  apply IHHbcd1 in H5...
  apply IHHbcd2 in H6...
  destruct H5 as [n1 Hl].
  destruct H6 as [n2 Hr].
  rewrite <- (app_nil_r [_]) in Hl.
  apply (wk_gen b) in Hl...
  destruct Hl as [m1 Hl].
  rewrite <- (app_nil_l [_]) in Hr.
  apply (wk_gen a) in Hr...
  destruct Hr as [m2 Hr].
  exists (S (S (S (max m1 m2)))).
  rewrite <- (app_nil_l [_]).
  apply il.
  apply ir ; eapply subtype_inc...
- inversion Hva.
  inversion Hvb.
  inversion H4.
  inversion H5.
  inversion H11.
  inversion H12.
  subst.
  apply IHHbcd1 in H23...
  apply IHHbcd2 in H19...
  destruct H23 as [n1 Hl].
  destruct H19 as [n2 Hr].
  exists (S (max n1 n2)).
  change [a → b] with
    (map (fun x => Some None @ fst x , snd x) [([a],[b])]).
  apply ct...
  + intros ; split ; simpl...
  + intros i Hi.
    constructor ; [ | constructor].
    inversion Hi ; [ | inversion H0] ; subst.
    eapply subtype_inc...
  + intros i Hi.
    inversion Hi ; [ | inversion H0] ; subst.
    eapply subtype_inc...
  + constructor ; [ | constructor ]...
    split ; [ | split ; [ | split]]...
- inversion Hvb.
  inversion H1.
  inversion H4.
  inversion H5.
  inversion H13.
  subst.
  assert (Ha := ax a H9).
  destruct Ha as [na Ha].
  assert (Hb := ax b H17).
  destruct Hb as [nb Hb].
  assert (Hc := ax c H18).
  destruct Hc as [nc Hc].
  rewrite <- (app_nil_r [_]) in Hb.
  apply (wk_gen c) in Hb...
  destruct Hb as [mb Hb].
  rewrite <- (app_nil_l [_]) in Hc.
  apply (wk_gen b) in Hc...
  destruct Hc as [mc Hc].
  eexists.
  rewrite <- (app_nil_l [_]).
  apply il.
  change [a → b ; a → c] with
    (map (fun x => Some None @ fst x , snd x) [([a],[b]);([a],[c])]).
  apply ct...
  + intros ; split...
  + intros i Hi.
    constructor ; [ | constructor ; [ | constructor] ].
    * inversion Hi ; [ | inversion H0] ; subst.
      apply (subtype_inc (S (S (max na (max mb mc)))))...
      eapply subtype_inc...
    * inversion Hi ; [ | inversion H0] ; subst.
      apply (subtype_inc (S (S (max na (max mb mc)))))...
      eapply subtype_inc...
  + intros i Hi.
    inversion Hi ; [ | inversion H0] ; subst ; simpl.
    apply ir ; eapply subtype_inc...
  + constructor ; [ | constructor ; [ | constructor]] ; 
      (split ; [ | split ; [ | split]])...
    * unfold trd ; constructor ; [ | constructor]...
    * unfold trd ; constructor ; [ | constructor]...
- eexists.
  rewrite <- (app_nil_l [_]).
  apply wk...
  change [] with (map (fun x => Some None @ fst x , snd x) []).
  apply ct ; intros ; try constructor...
  + constructor ; [ | constructor]...
  + inversion H ; [ | inversion H1] ; simpl.
    change [] with (map (fun x => None @ fst x , snd x) []).
    apply ct ; intros ; try constructor...
    * constructor.
    * inversion H0.
  Unshelve. exact 0.
- assert (Forall (fun x => exists k, subtype [x] x k) l1) as Hl1.
  { inversion Hva ; subst.
    assert (Forall ivalid l1) as Hv1.
    { revert H5 ; clear ; induction l1 ; intro Hv ; constructor ; inversion Hv...
      apply IHl1... }
    revert Hv1 ; clear ; induction l1 ; constructor ; inversion Hv1.
    - apply ax...
    - apply IHl1... }
  apply AFClist in Hl1 ; [ | intros ; eapply subtype_inc ; eassumption].
  destruct Hl1 as [n1 Hl1].
  assert (Forall (fun x => exists k, subtype [x] x k) l2) as Hl2.
  { inversion Hva ; subst.
    assert (Forall ivalid l2) as Hv2.
    { revert H5 ; clear ; induction l1 ; intro Hv ; inversion Hv...
      apply IHl1... }
    revert Hv2 ; clear ; induction l2 ; constructor ; inversion Hv2.
    - apply ax...
    - apply IHl2... }
  apply AFClist in Hl2 ; [ | intros ; eapply subtype_inc ; eassumption].
  destruct Hl2 as [n2 Hl2].
  assert (ivalid a) as Hva1.
  { inversion Hva ; subst.
    revert H5 ; clear ; induction l1 ; intro Hva ; inversion Hva ; subst...
    apply IHl1... }
  assert (ivalid b) as Hvb1.
  { inversion Hvb ; subst.
    revert H5 ; clear ; induction l1 ; intro Hvb ; inversion Hvb ; subst...
    apply IHl1... }
  apply IHHbcd in Hva1...
  destruct Hva1 as [n Hab].
  exists (S (max n (max n1 n2))).
  change [Some (Some C) @ [] , (l1 ++ a :: l2)] with
    (map (fun x => Some (Some C) @ fst x , snd x) [([],(l1 ++ a :: l2))]).
  apply ct...
  + intros Hlen ; inversion Hlen.
  + intros i Hi.
    inversion Hi.
  + intros i Hi.
    unfold trd ; simpl.
    assert (i < length (l1 ++ a :: l2)) as Hv.
    { inversion Hva ; subst... }
    clear Hvb Hva.
    revert l1 Hv Hab Hl1 Hl2 ; clear ; induction i ; intros l1 Hv Hab Hl1 Hl2 ;
      destruct l1 ; simpl...
    * eapply subtype_inc...
    * inversion Hl1.
      eapply subtype_inc...
    * eapply subtype_inc.
      apply Forall_nth...
      simpl in Hv...
      idtac...
    * apply IHi...
      simpl in Hv...
      inversion Hl1...
  + constructor ; [ | constructor] ; simpl...
    split ; [ | split ; [ | split]] ;  inversion Hva...
  + inversion Hva...
    rewrite <- H3.
    clear ; induction l1 ; simpl...
- assert (forall l l', Forall ivalid l -> Forall ivalid l' -> 
    Forall (fun x => exists n, Forall (fun y => @subtype BCDC [y;x] x n) l') l)
  as HAXl.
  { clear ; induction l ; constructor.
    - apply AFClist.
      + intros.
        eapply subtype_inc...
      + revert H0 ; induction l' ; intros Hvl' ; constructor.
        * inversion H.
          assert (Hax := ax a H2).
          destruct Hax as [n Hax].
          rewrite <- (app_nil_l [a0; a]).
          eapply wk_gen...
          inversion Hvl'...
        * inversion Hvl'.
          apply IHl'...
    - apply IHl...
      inversion H... }
  inversion Hva.
  inversion H1.
  inversion H2 ; subst.
  assert (HAX1 := HAXl l1 l2 H9).
  apply AFClist in HAX1...
  destruct HAX1 as [n1 HAX1].
  assert (HAX2 := HAXl l2 l1 H16).
  apply AFClist in HAX2...
  destruct HAX2 as [n2 HAX2].
  exists (S (S (S (S (max n1 n2))))).
  rewrite <- (app_nil_l [_]).
  apply il.
  change [Some (Some C) @ [] , l1 ; Some (Some C) @ [] , l2]
    with (map (fun x => Some (Some C) @ fst x , snd x) [([],l1);([],l2)]).
  apply ct...
  + intros Hl.
    inversion Hl.
  + intros i Hi.
    inversion Hi.
  + intros i Hi.
    unfold trd ; simpl.
    rewrite <- H7 in Hi.
    rewrite <- H7 in H14.
    assert (Forall ivalid l1) as Hv1.
    { inversion Hva ; inversion H3... }
    assert (Forall ivalid l2) as Hv2.
    { inversion Hva ; inversion H4... }
    clear Hva Hvb H1 H2 H7 H9 H16.
    revert l1 l2 Hv1 Hv2 HAX1 HAX2 H14 Hi ; induction i ;
      intros l1 l2 Hv1 Hv2 HAX1 HAX2 Heql Hi...
    * destruct l1 ; try (inversion Hi ; fail).
      destruct l2 ; inversion Heql.
      simpl.
      inversion HAX1.
      inversion HAX2.
      inversion H2.
      inversion H7.
      apply ir.
      eapply ex.
      eapply subtype_inc...
      apply Permutation_rev.
      eapply subtype_inc...
    * destruct l1 ; try (inversion Hi ; fail).
      destruct l2 ; inversion Heql ; simpl.
      apply IHi...
      -- inversion Hv1...
      -- inversion Hv2...
      -- inversion HAX1...
         revert H3 ; clear ; induction l1 ; intros HF1 ; constructor ; inversion HF1.
         ++ inversion H1...
         ++ apply IHl1...
      -- inversion HAX2...
         revert H3 ; clear ; induction l2 ; intros HF2 ; constructor ; inversion HF2.
         ++ inversion H1...
         ++ apply IHl2...
      -- simpl in Hi...
  + constructor ; constructor ; constructor ; try constructor...
    split.
    * unfold scd ; simpl ; constructor.
    * unfold trd ; simpl...
  + inversion Hvb...
  + intros.
    eapply (inc_Forall (fun z y => subtype [y;a] a z))...
    intros.
    eapply subtype_inc...
  + intros.
    eapply (inc_Forall (fun z y => subtype [y;a] a z))...
    intros.
    eapply subtype_inc...
Qed.


(** * Completeness with respect to BCD *)
Fixpoint interlist l :=
  match l with
  | [] => Ω
  | a::r => a ∩ (interlist r)
  end.

Lemma ivalid_interlist : forall l,
  Forall ivalid l -> ivalid (interlist l).
Proof with simpl ; try reflexivity ; try assumption ; try omega.
induction l ; intro Hv.
- constructor ; try constructor...
- inversion Hv.
  constructor...
  apply IHl...
Qed.

Lemma lvalid_interlist : forall C L, lvalid C L ->
  ivalid (interlist ((map (fun c => (C@fst c,snd c)) L))).
Proof with simpl ; try assumption ; try omega.
induction L ; intros Hvl ; constructor ; try (constructor ; fail)...
- inversion Hvl.
  destruct H1 as (Hv1 & Hv2 & Hv3 & Hv4).
  constructor...
- apply IHL...
  inversion Hvl...
Qed.

Lemma interlist_app : forall G D, Forall ivalid (G ++ D) ->
  interlist (G ++ D) ≤ interlist G ∩ interlist D.
Proof with try assumption.
induction G ; intros D Hv.
- eapply trans_bcd.
  + apply inter_right_bcd.
  + apply inter_bcd.
    * apply omega_right_bcd.
    * apply id_bcd.
  + constructor ; apply ivalid_interlist...
- eapply trans_bcd.
  + rewrite <- app_comm_cons.
    eapply trans_bcd.
    * apply inter_bcd.
      apply id_bcd.
      apply IHG.
      inversion Hv...
    * eapply trans_bcd.
      apply inter_right_bcd.
      apply inter_bcd.
      apply inter_bcd.
      apply id_bcd.
      apply inter_left1_bcd.
      apply inter_left2_bcd.
      inversion Hv.
      apply Forall_app_inv in H2.
      destruct H2.
      constructor ; constructor ; try constructor ;
        try apply ivalid_interlist...
    * inversion Hv.
      apply Forall_app_inv in H2.
      destruct H2.
      constructor ; try constructor ; try apply ivalid_interlist...
  + apply inter_bcd.
    * apply id_bcd.
    * apply inter_left2_bcd.
  + inversion Hv.
      apply Forall_app_inv in H2.
      destruct H2.
      constructor ; constructor ; try constructor ;
        try apply ivalid_interlist...
Qed.

Lemma interlist_app_inv : forall G D, Forall ivalid (G ++ D) ->
  interlist G ∩ interlist D ≤ interlist (G ++ D).
Proof with try assumption.
induction G ; intros D Hv.
- apply inter_left2_bcd.
- inversion Hv.
  apply Forall_app_inv in H2.
  destruct H2.
  eapply trans_bcd.
  + eapply trans_bcd.
    * apply inter_right_bcd.
    * apply inter_bcd.
      eapply trans_bcd.
      apply inter_left1_bcd.
      apply inter_left1_bcd.
      constructor ; try constructor ; try constructor ;
        try apply ivalid_interlist...
      apply inter_bcd.
      apply inter_left2_bcd.
      apply id_bcd.
    * constructor ; try constructor ; try constructor ;
        try apply ivalid_interlist...
  + rewrite <- app_comm_cons.
    apply inter_bcd.
    apply id_bcd.
    apply IHG.
    inversion Hv...
  + constructor ; try constructor ; try constructor ;
      try apply ivalid_interlist...
Qed.

Theorem completeness : forall G s n, subtype G s n -> interlist G ≤ s.
Proof with try eassumption ; try reflexivity ; try (zify ; omega).
intros G s n ; revert G s.
induction n using (well_founded_induction lt_wf) ;
  intros G s pi ; inversion pi ; subst.
- apply subtype_ivalid in pi.
  destruct pi as [Hv _].
  apply Forall_app_inv in Hv.
  destruct Hv as [Hv1 Hv2].
  inversion Hv2 ; subst.
  assert (Hv3 := Forall_app _ _ _ Hv1 H5).
  apply H in H0...
  eapply trans_bcd.
  + apply interlist_app.
    apply Forall_app...
  + eapply trans_bcd.
    * apply inter_bcd.
      apply id_bcd.
      apply inter_left2_bcd.
    * eapply trans_bcd.
      apply interlist_app_inv.
      assumption.
      assumption.
      apply ivalid_interlist...
    * constructor ; apply ivalid_interlist...
  + constructor ; try constructor ; try apply ivalid_interlist...
- eapply trans_bcd.
  + apply inter_right_bcd.
  + apply inter_bcd.
    * apply H in H0...
    * apply H in H1...
  + apply subtype_ivalid in H1.
    destruct H1 as [Hv _].
    apply ivalid_interlist in Hv.
    constructor...
- apply subtype_ivalid in pi.
  destruct pi as [Hv _].
  apply Forall_app_inv in Hv.
  destruct Hv as [Hv1 Hv2].
  inversion Hv2 ; subst.
  assert (Hv3 := Forall_app _ _ _ Hv1 H4).
  inversion H3 ; subst.
  apply H in H0...
  eapply trans_bcd.
  + apply interlist_app.
    apply Forall_app...
  + eapply trans_bcd.
    * apply inter_bcd.
      apply id_bcd.
      eapply trans_bcd.
      apply inter_right_bcd.
      apply inter_bcd.
      eapply trans_bcd.
      apply inter_left1_bcd.
      apply inter_left1_bcd.
      assumption.
      simpl ; apply inter_bcd.
      apply inter_left2_bcd.
      apply id_bcd.
      constructor ; constructor ; try apply ivalid_interlist...
    * eapply trans_bcd.
      change (s0 ∩ (t ∩ (interlist D))) with (interlist (s0 :: t :: D)).
      apply interlist_app_inv.
      apply Forall_app...
      constructor ; try constructor...
      assumption.
      apply ivalid_interlist.
      apply Forall_app...
      constructor ; try constructor...
    * constructor ; try constructor ; try constructor ;
        try apply ivalid_interlist...
  + constructor ; try constructor ; try apply ivalid_interlist...
- destruct C ; [ destruct o | ].
  + destruct a ; inversion H4.
    assert (forall i, i < β (Some (Some b0)) ->
              interlist (map (fun c => nth i (snd c) X) L) ≤ nth i b X) as Hb.
    { intros i Hi.
      apply H2 in Hi...
      apply H in Hi... }
    apply subtype_ivalid in pi.
    destruct pi as [_ Hvb].
    remember (length L) as k.
    clear H1 H2.
    revert L Heqk H0 Hb H3 Hvb H ; clear ;
      induction k using (well_founded_induction lt_wf) ;
      intros L Heqk He Hb Hlv Hvb IH.
    destruct k ; [ | destruct k] ; destruct L ; inversion Heqk.
    * assert (HF := He eq_refl).
      destruct HF as [HF _].
      simpl in HF.
      inversion HF.
    * destruct L ; inversion Heqk.
      inversion Hlv ; subst.
      destruct H3 as (Hlv1 & Hlv2 & Hlv3 & Hlv4) ; simpl.
      remember (fst p) as a.
      destruct a ; inversion Hlv1.
      remember (snd p) as b'.
      change (Ctor BCDC) with BCDCtor in Heqa.
      rewrite_all <- Heqa.
      change (Ctor BCDC) with BCDCtor in Heqb'.
      rewrite_all <- Heqb'.
      eapply trans_bcd ; [ | eapply trans_bcd | ].
      -- apply inter_left1_bcd.
      -- inversion Hvb.
         apply (ctor_mon_bcd_gen _ _ b) ; try constructor...
         simpl in Hb.
         simpl in H6.
         assert (length b' = bvar b0) as Hlen'.
         { simpl in Hlv2.
           rewrite Hlv2... }
         assert (length b' = length b) as Hlen by omega.
         rewrite <- Heqb' in Hb.
         rewrite <- Hlen' in Hb.
         clear H Hvb H3 H6.
         revert b Hlen Hb Hlv4 H8 ; clear ; induction b' ; intros b Hlen Hb Hvb' Hvb.
         destruct b ; simpl ; try constructor.
         inversion Hlen.
         inversion Hvb'.
         destruct b ; simpl ; try constructor.
         ++ inversion Hlen.
         ++ assert (0 < length (a :: b')) as Hb0 by (simpl ; omega).
            apply Hb in Hb0.
            eapply trans_bcd ; [ | eapply trans_bcd | ].
            ** apply inter_right_bcd.
            ** apply inter_bcd.
               --- apply id_bcd.
               --- apply omega_right_bcd.
            ** simpl in Hb0...
            ** constructor ; try constructor ; try constructor...
            ** constructor...
         ++ inversion Hvb.
            apply IHb'...
            ** inversion Hlen...
            ** intros i0 Hi0.
               apply (Hb (S i0)).
               simpl...
      -- apply id_bcd.
      -- inversion Hvb.
         constructor...
      -- constructor...
    * destruct L ; inversion Heqk.
      simpl.
      inversion Hlv ; subst.
      destruct H4 as (Hlv1 & Hlv2 & Hlv3 & Hlv4).
      remember (fst p) as a1.
      destruct a1 ; inversion Hlv1.
      remember (snd p) as b1'.
      inversion H5.
      destruct H3 as (Hlv1' & Hlv2' & Hlv3' & Hlv4').
      remember (fst p0) as a2.
      destruct a2 ; inversion Hlv1'.
      remember (snd p0) as b2'.
      change (Ctor BCDC) with BCDCtor in Heqa1.
      rewrite_all <- Heqa1.
      change (Ctor BCDC) with BCDCtor in Heqb1'.
      rewrite_all <- Heqb1'.
      change (Ctor BCDC) with BCDCtor in Heqa2.
      rewrite_all <- Heqa2.
      change (Ctor BCDC) with BCDCtor in Heqb2'.
      rewrite_all <- Heqb2'.
      apply lvalid_interlist in H4.
      assert (length b1' = bvar b0) as Hlen1.
      { simpl in Hlv2.
        rewrite Hlv2... }
      assert (length b2' = bvar b0) as Hlen2.
      { simpl in Hlv2'.
        rewrite Hlv2'... }
      assert (length b1' = length b2') as Hlen by omega.
      assert (Forall ivalid (map2 inter b1' b2')) as Hvinter.
      { clear Heqb2' Hlv2' Hlen2.
        revert b2' Hlen Hlv4 Hlv4' ; clear ; induction b1' ; intros b2' Hlen H1 H2.
        simpl ; constructor.
        destruct b2'.
        simpl ; constructor.
        inversion H1.
        inversion H2.
        simpl in Hlen.
        constructor ; try constructor...
        apply IHb1'... }
      apply (trans_bcd _ (((κ b0 b1') ∩ (κ b0 b2'))
                          ∩ interlist (map (fun c => Some (Some b0) @ fst c, snd c) L))) ;
        [ eapply trans_bcd | | ].
      -- apply inter_right_bcd.
      -- apply inter_bcd.
         ++ apply inter_bcd.
            ** apply id_bcd.
            ** eapply trans_bcd.
               --- apply inter_left1_bcd.
               --- apply id_bcd.
               --- constructor...
         ++ eapply trans_bcd.
            ** apply inter_left2_bcd.
            ** apply inter_left2_bcd.
            ** constructor ; try constructor...
      -- constructor ; constructor ; constructor ; try constructor...
      -- apply (trans_bcd _ ((κ b0 (map2 inter b1' b2'))
                          ∩ interlist (map (fun c => Some (Some b0) @ fst c, snd c) L))).
         ++ apply inter_bcd.
            ** apply ctor_inter_bcd.
            ** apply id_bcd.
         ++ change ((Some (Some b0) @ [], map2 inter b1' b2') ∩
                       interlist (map (fun c => Some (Some b0) @ fst c, snd c) L))
              with (interlist (map (fun c => Some (Some b0) @ fst c, snd c)
                       (([],map2 inter b1' b2') :: L))).
            apply (H (S (length L)))...
            ** simpl in Hb.
               rewrite <- Heqb1' in Hb.
               rewrite <- Heqb2' in Hb.
               simpl in Hb ; simpl.
               assert (forall i, i < bvar b0 ->
                         ivalid (interlist (map (fun c => nth i (snd c) X) L))) as HvL.
               { revert H4 H5 ; clear ; induction L ; intros Hv HC i Hi ;
                   constructor ; try constructor...
                 - inversion Hv.
                   inversion H1 ; subst.
                   eapply Forall_nth...
                   change BCDCtor with (Ctor BCDC) in H7.
                   rewrite H7.
                   simpl...
                 - inversion Hv.
                   eapply IHL in H2...
                   apply H2.
                   inversion HC.
                   inversion H6.
                   constructor... }
               intros i Hi.
               assert (ivalid (nth i (map2 inter b1' b2') X)) as Hvi.
               { rewrite (nth_map2 _ _ _ _ X X).
                 constructor ; try constructor...
                 apply Forall_nth...
                 apply Forall_nth...
                 apply (map2_length inter) in Hlen... }
               apply (trans_bcd _ ((nth i b1' X) ∩ ((nth i b2' X) ∩
                                interlist (map (fun c => nth i (snd c) X) L)))) ;
                 [ | apply Hb | ]...
               --- eapply trans_bcd.
                   +++ apply inter_right_bcd.
                   +++ apply inter_bcd.
                       *** eapply trans_bcd...
                           ---- apply inter_left1_bcd.
                           ---- rewrite (nth_map2 _ _ _ _ X X).
                                ++++ apply inter_left1_bcd.
                                ++++ apply (map2_length inter) in Hlen...
                       *** apply inter_bcd.
                           ---- rewrite (nth_map2 _ _ _ _ X X).
                                ++++ apply inter_left2_bcd.
                                ++++ apply (map2_length inter) in Hlen...
                           ---- apply id_bcd.
                   +++ constructor ; constructor ; try constructor...
                       *** apply HvL...
                       *** apply HvL...
               --- constructor ; try constructor ; try constructor...
                   +++ apply Forall_nth...
                   +++ apply Forall_nth...
                   +++ apply HvL...
            ** constructor.
               --- split ; [ | split ; [ | split]] ; simpl...
                   apply (map2_length inter) in Hlen...
               --- inversion Hlv.
                   inversion H8...
         ++ constructor ; try constructor ; try constructor ; try constructor...
            apply (map2_length inter) in Hlen.
            simpl...
      -- constructor ; try constructor ; try constructor ; try constructor...
  + destruct a ; inversion H4.
    destruct a ; inversion H7.
    destruct b ; inversion H5.
    destruct b ; inversion H8.
    assert (ivalid i) as Hvi.
    { assert (0 < 1) as HS by omega.
      assert (Hvi := H1 _ HS).
      inversion Hvi.
      - rewrite <- H6 in H0.
        assert (Hi := H0 eq_refl).
        destruct Hi.
        inversion H10...
      - apply subtype_ivalid in H6.
        destruct H6.
        inversion H6... }
    assert (Forall (fun c => i ≤ (nth 0 (fst c) X)) L) as Ha.
    { revert H1 ; clear - H ; induction L ; intro Ha ; constructor.
      - assert (0 < 1) as HS by omega.
        apply Ha in HS.
        inversion HS...
        assert (Hvi := H2).
        apply subtype_ivalid in Hvi.
        destruct Hvi as [Hvi _].
        inversion Hvi.
        eapply trans_bcd.
        eapply trans_bcd.
        apply inter_right_bcd.
        apply inter_bcd.
        apply id_bcd.
        apply omega_right_bcd.
        constructor...
        change (i ∩ Ω) with (interlist [i]).
        eapply (H _ (le_n _) _ _ H2)...
        constructor ; try constructor...
      - apply IHL...
        intros i0 Hi0.
        apply Ha in Hi0.
        inversion Hi0... }
    assert (interlist (map (fun c => nth 0 (snd c) X) L) ≤ i0) as Hb.
    { assert (0 < 1) as HS by omega.
      apply H2 in HS...
      apply H in HS... }
    remember (length L) as k.
    clear pi H1 H2.
    revert L Heqk H0 Ha Hb H3 H Hvi ; clear ;
      induction k using (well_founded_induction lt_wf) ;
      intros L Heqk He Ha Hb Hlv IH Hvi.
    destruct k ; [ | destruct k] ; destruct L ; inversion Heqk.
    * apply (trans_bcd _ (i → Ω)).
      -- apply (trans_bcd _ (Ω → Ω)).
         ++ apply arrow_omega_bcd.
         ++ apply arrow_bcd.
            ** apply omega_right_bcd.
            ** apply omega_right_bcd.
         ++ constructor ; constructor ; constructor ; constructor...
      -- apply arrow_bcd...
         apply id_bcd.
      -- constructor ; constructor ; try constructor ; try constructor...
    * destruct L ; inversion Heqk.
      inversion Hlv ; subst.
      destruct H3 as (Hlv1 & Hlv2 & Hlv3 & Hlv4) ; simpl.
      remember (fst p) as a.
      destruct a ; inversion Hlv1.
      destruct a ; inversion H2.
      remember (snd p) as b.
      destruct b ; inversion Hlv2.
      destruct b ; inversion H3.
      change (Ctor BCDC) with BCDCtor in Heqa.
      rewrite_all <- Heqa.
      change (Ctor BCDC) with BCDCtor in Heqb.
      rewrite_all <- Heqb.
      inversion Hlv3.
      inversion Hlv4.
      eapply trans_bcd.
      -- apply inter_left1_bcd.
      -- apply arrow_bcd.
         ++ inversion Ha.
            change BCDCtor with (Ctor BCDC) in Heqa.
            rewrite <- Heqa in H14...
         ++ simpl in Hb.
            rewrite <- Heqb in Hb.
            simpl in Hb.
            eapply trans_bcd ; [ eapply trans_bcd | | ].
            ** apply inter_right_bcd.
            ** apply inter_bcd.
               --- apply id_bcd.
               --- apply omega_right_bcd.
            ** constructor...
            ** apply Hb.
            ** constructor ; try constructor...
      -- constructor...
    * destruct L ; inversion Heqk.
      simpl.
      inversion Hlv ; subst.
      destruct H4 as (Hlv1 & Hlv2 & Hlv3 & Hlv4).
      remember (fst p) as a1.
      destruct a1 ; inversion Hlv1.
      destruct a1 ; inversion H2.
      remember (snd p) as b1.
      destruct b1 ; inversion Hlv2.
      destruct b1 ; inversion H3.
      inversion H5 ; subst.
      destruct H6 as (Hlv1' & Hlv2' & Hlv3' & Hlv4').
      remember (fst p0) as a2.
      destruct a2 ; inversion Hlv1'.
      destruct a2 ; inversion H4.
      remember (snd p0) as b2.
      destruct b2 ; inversion Hlv2'.
      destruct b2 ; inversion H6.
      change (Ctor BCDC) with BCDCtor in Heqa1.
      rewrite_all <- Heqa1.
      change (Ctor BCDC) with BCDCtor in Heqb1.
      rewrite_all <- Heqb1.
      change (Ctor BCDC) with BCDCtor in Heqa2.
      rewrite_all <- Heqa2.
      change (Ctor BCDC) with BCDCtor in Heqb2.
      rewrite_all <- Heqb2.
      apply lvalid_interlist in H7.
      inversion Hlv3.
      inversion Hlv4.
      inversion Hlv3'.
      inversion Hlv4'.
      apply (trans_bcd _ (((i1 → i2) ∩ (i3 → i4))
                          ∩ interlist (map (fun c => Some None @ fst c, snd c) L))).
      -- eapply trans_bcd.
         ++ apply inter_right_bcd.
         ++ apply inter_bcd.
            ** apply inter_bcd.
               --- apply id_bcd.
               --- apply inter_left1_bcd.
            ** eapply trans_bcd.
               --- apply inter_left2_bcd.
               --- apply inter_left2_bcd.
               --- constructor ; try constructor...
         ++ constructor ; constructor ; constructor ; try constructor...
      -- apply (trans_bcd _ ((i1 ∩ i3 → i2 ∩ i4)
                          ∩ interlist (map (fun c => Some None @ fst c, snd c) L))).
         ++ apply inter_bcd.
            ** apply (trans_bcd _ ((i1 ∩ i3 → i2) ∩ (i1 ∩ i3 → i4))).
               --- apply inter_bcd.
                   +++ apply arrow_bcd.
                       *** apply inter_left1_bcd.
                       *** apply id_bcd.
                   +++ apply arrow_bcd.
                       *** apply inter_left2_bcd.
                       *** apply id_bcd.
               --- eapply trans_bcd.
                   +++ apply arrow_inter_bcd.
                   +++ apply arrow_bcd ; apply id_bcd.
                   +++ constructor ; constructor ; constructor...
               --- constructor ; constructor ; constructor ; try constructor...
            ** apply id_bcd.
         ++ change ((i1 ∩ i3 → i2 ∩ i4) ∩
                      interlist (map (fun c => Some None @ fst c, snd c) L))
            with (interlist (map (fun c => Some None @ fst c, snd c)
                      (([i1 ∩ i3],[i2 ∩ i4])::L))).
            apply (H (S (length L)))...
            ** inversion Ha.
               inversion H26.
               constructor...
               eapply trans_bcd.
               --- apply inter_right_bcd.
               --- apply inter_bcd.
                   +++ change  BCDCtor with (Ctor BCDC) in Heqa1.
                       rewrite <- Heqa1 in H25.
                       apply H25.
                   +++ change BCDCtor with (Ctor BCDC) in Heqa2.
                       rewrite <- Heqa2 in H29.
                       apply H29.
               --- constructor...
            ** simpl in Hb.
               rewrite <- Heqb1 in Hb.
               rewrite <- Heqb2 in Hb.
               simpl in Hb ; simpl.
               assert (ivalid (interlist (map (fun c => nth 0 (snd c) X) L))) as HvL.
               { revert H7 ; clear ; induction L ; intros Hv ;
                 constructor ; try constructor...
                 - inversion Hv ; subst.
                   inversion H1.
                   apply Forall_nth...
                   change BCDCtor with (Ctor BCDC) in H5 ; rewrite H5.
                   simpl...
                 - inversion Hv.
                   apply IHL... }
               apply (trans_bcd _
                      (i2 ∩ (i4 ∩ interlist (map (fun c => nth 0 (snd c) X) L))))...
               --- eapply trans_bcd.
                   +++ apply inter_right_bcd.
                   +++ apply inter_bcd.
                       *** eapply trans_bcd.
                           ---- apply inter_left1_bcd.
                           ---- apply inter_left1_bcd.
                           ---- constructor...
                       *** apply inter_bcd.
                           ---- apply inter_left2_bcd.
                           ---- apply id_bcd.
                   +++ constructor ; constructor ; try constructor...
               --- constructor ; try constructor ; try constructor...
            ** constructor.
               --- split ; [ | split ; [ | split]] ; simpl...
                   +++ constructor ; constructor...
                   +++ constructor ; constructor...
               --- inversion Hlv.
                   inversion H26...
          ++ constructor ; try constructor ; try constructor ; try constructor...
       -- constructor ; try constructor ; try constructor ; try constructor...
  + destruct a ; inversion H4.
    destruct b ; inversion H5.
    apply omega_right_bcd.
Qed.

End BCDstd.


